importPackage(java.util);
importPackage(java.io);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLMessage);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLToken);
importClass(Packages.com.thoughtworks.xstream.XStream);
importClass(Packages.com.thoughtworks.xstream.converters.reflection.PureJavaReflectionProvider);
importClass(Packages.com.alarmpoint.integrationagent.security.EncryptionUtils);
importClass(Packages.com.alarmpoint.integrationagent.util.EventDeduplicator);

var xStream = new XStream(new PureJavaReflectionProvider())

var IA_UTIL_VERSION = "lib_1_1_0";

load("integrationservices/caservicedesk/lib/" + IA_UTIL_VERSION + "/javascript/core/baseclass.js");
load("integrationservices/caservicedesk/lib/" + IA_UTIL_VERSION + "/javascript/core/logger.js");
load("integrationservices/caservicedesk/lib/" + IA_UTIL_VERSION + "/javascript/core/util.js");
load("integrationservices/caservicedesk/lib/" + IA_UTIL_VERSION + "/javascript/webservices/wsutil.js");
load("integrationservices/caservicedesk/lib/" + IA_UTIL_VERSION + "/javascript/xmatters/xmattersws.js");
load("integrationservices/caservicedesk/lib/" + IA_UTIL_VERSION + "/javascript/xmatters/xMattersUserSync.js");
load("integrationservices/caservicedesk/lib/" + IA_UTIL_VERSION + "/javascript/xmatters/xMattersGroupSync.js");
load("integrationservices/caservicedesk/lib/" + IA_UTIL_VERSION + "/javascript/xmatters/dataSyncHelper.js");
load("integrationservices/caservicedesk/lib/" + IA_UTIL_VERSION + "/javascript/xmatters/dataSyncDevice.js");
load("integrationservices/caservicedesk/lib/" + IA_UTIL_VERSION + "/javascript/xmatters/dataSyncUser.js");
load("integrationservices/caservicedesk/lib/" + IA_UTIL_VERSION + "/javascript/xmatters/dataSyncGroup.js");
load("integrationservices/caservicedesk/lib/" + IA_UTIL_VERSION + "/javascript/xmatters/dataSyncGroupMember.js");
load("integrationservices/caservicedesk/lib/" + IA_UTIL_VERSION + "/javascript/xmatters/dataSyncTeam.js");
load("integrationservices/caservicedesk/lib/" + IA_UTIL_VERSION + "/javascript/xmatters/dataSyncCoverage.js");

load("integrationservices/caservicedesk/casddataload/configuration.js");
load("integrationservices/caservicedesk/casddataload/dataSyncList.js");
load("integrationservices/caservicedesk/casddataload/caservicedeskws.js");
load("integrationservices/caservicedesk/casddataload/processUsers.js");
load("integrationservices/caservicedesk/casddataload/processGroups.js");
load("integrationservices/caservicedesk/casddataload/supervisorMapper.js");
load("integrationservices/caservicedesk/casddataload/userRolesMapper.js");
load("integrationservices/caservicedesk/casddataload/phoneNumber.js");

// The location of the soapUI project file that is used to access xMatters Web services depends on the integration
// that it's being used in, so set the path here.
xMattersWS.prototype.XMATTERS_WSDL_DEFN = "integrationservices/caservicedesk/lib/" + IA_UTIL_VERSION + "/resources/soap/xmatters/v4.1/AlarmPoint-soapui-project.xml";

var exceptionString = "JavaException: ";

var notePrefix = "[xMatters] - "; // The Prefix which is added to the note comments for xMatters e.g. "[xMatters] " 

var TOKEN_NAME = "apiaResult";
var TOKEN_NAME_EXCEPTION = "apiaResultException";

/**
 * This is the main method for handling APXML messages sent to the
 * Integration Service by APClient.  APClient sends the location of the CA Service Desk notification file.
 * This notification file should be created as part of the xMatters notification method.  The file is then
 * parsed for parameters which will be sent to AlarmPoint as event tokens.
 * <p>
 * For example, the key NX_NTF_MESSAGE_HTML is found in the notification file and is sent to AlarmPoint
 * as NOTIFICATION_MESSAGE_HTML_0 , NOTIFICATION_MESSAGE_HTML_1.  It is broken up as each event token is
 * limited to 4000k.  Once recieved by AlarmPoint these tokens are re-assembled and sent as an html notification
 * to the targeted recipient.
 * <p>
 * Any APXMLMessage object that this method returns will be sent to AlarmPoint
 * via the Integration Service's outbound queues.
 */
function apia_input(apxml)
{
    log.info("Enter - apia_input");

    // The functions that process the request need the requests action and CA SDM object id extracted from the AP XML.
    var syncRequest =
    {
        action : apxml.getValue("action")
    };

    if (isEmpty(syncRequest.action))
    {
        throw "No action given";
    }

    var caServiceDeskWS = new CAServiceDeskWS(new WSUtil(), SERVICE_DESK_URL);

    // Used to record any exception encountered during the data load operation that causes the overall operation
    // to abort (as opposed to local problems with a user or device that do not terminate the operation
    var syncAbortException = null;

    try
    {
        // Reset the success/warning/failure collections
        failedList = new ArrayList();
        successList = new ArrayList();
        warningList = new ArrayList();
        reprocessedList = new ArrayList();

        caServiceDeskWS.login(SERVICE_DESK_USER, SERVICE_DESK_PASSWORD);

        processDataLoadRequest(caServiceDeskWS, syncRequest);
    }
    catch (e)
    {
        syncAbortException = e;
    }
    finally
    {
        logSyncResults(syncRequest, syncAbortException);
        caServiceDeskWS.logout();
    }

    log.info("Exit - apia_input");
}


function processDataLoadRequest(caServiceDeskWS, requestObject)
{
    log.info("Enter - processDataLoadRequest");

    log.info("processDataLoadRequest: processing request - action [" + requestObject.action + "]");

    switch (String(requestObject.action.toLowerCase()))
    {
        case ACTION_LOAD:
            log.info("processDataLoadRequest: starting batch data load");
            processCASDUsers(caServiceDeskWS, makeUserBatchWhereClause(), PROCESS_ADD_UPDATE, ALLOW_ADD_USER);
            processCASDGroups(caServiceDeskWS, makeGroupBatchWhereClause(), PROCESS_ADD_UPDATE, ALLOW_ADD_GROUP);
            break;

        // This integration only supports batch data load

        default:
            throw "Request action could not be matched to an operation";
    }

    log.info("Exit - processDataLoadRequest");
}

//----------------------------------------------------------------------------------------------------------------------
// Utility functions for constructing the "where clause" for the Web service operations that get lists of users and groups.
//----------------------------------------------------------------------------------------------------------------------

function makeUserBatchWhereClause()
{
    return WHERE_CLAUSE_USERS + makeWhereClauseForSyncList("users", "User", "userid");
}

function makeGroupBatchWhereClause()
{
    return WHERE_CLAUSE_GROUPS + makeWhereClauseForSyncList("groups", "Group", "last_name");
}

function makeGroupMembersWhereClause(groupId)
{
    return WHERE_CLAUSE_GROUP_MEMBERS.replace("${GROUP_ID}", groupId) + makeWhereClauseForSyncList("users", "User", "member.userid");
}


/**
 * Creates the part of the Web Service where clause that reflects the syncList and
 * either prevents excluded users/groups being retrieved or specifically selects included
 * users/groups.
 *
 * @param nodeName name of the parent node containing the ids that are required
 * @param type user or group (to retrieve the correct data from the list)
 * @param fieldName attribute name used to identify the data
 */
function makeWhereClauseForSyncList(nodeName, type, fieldName)
{
    log.info("makeWhereClauseForSyncList - nodeName [" + nodeName + "], type [" + type + "], fieldName [" + fieldName + "]");

    var whereClause = new java.lang.StringBuffer("");

    var exclude = SYNC_ACTION.equalsIgnoreCase("exclude");
    var operator = exclude ? "!=" : "=";
    var concatenator = exclude ? " AND " : " OR ";

    var usersOrGroupElement = syncList.child(nodeName);

    log.debug("makeWhereClauseForSyncList - usersOrGroupElement = [ " + usersOrGroupElement + "]");

    if ((usersOrGroupElement != null) && (usersOrGroupElement.child(type.toLowerCase()).length() > 0))
    {
        whereClause.append(" AND (");

        // Iterate the users/groups in the selected node and build up the where clause
        for (var i = 0; i < usersOrGroupElement.child(type.toLowerCase()).length(); i++)
        {
            var name = usersOrGroupElement.child(type.toLowerCase())[i];
            whereClause.append(((i > 0) ? concatenator : ""));
            whereClause.append(fieldName + " " + operator + " '" + name.toString().replace("'", "''") + "'");

            log.info(type + ": " + name + " will " + (exclude ? "NOT" : "") + " be synchronized as " + name + " is on the " + (exclude ? "EXCLUDE" : "INCLUDE") + " list");
        }

        whereClause.append(")");
    }
    else if (!exclude)
    {
        // An empty sync list when the sync action is include is ignored as it is assumed the user wouldn't want to
        // include no users and groups in this manner.

        log.warn("makeWhereClauseForSyncList - SYNC_ACTION is include and synclist for [" + nodeName + "] is empty. Ignoring sync list during WHERE clause construction.");
    }

    var result = whereClause.toString();

    log.debug("makeWhereClauseForSyncList - where clause = [ " + result + "]");
    log.info("Exit - makeWhereClauseForSyncList");
    return result;
}

//----------------------------------------------------------------------------------------------------------------------
// Utility functions for logging the results of data load requests, and sending a notification to a specific user
// via xMatters.
//----------------------------------------------------------------------------------------------------------------------

function logSyncResults(syncRequest, syncAbortException)
{
    log.info("Enter - logSyncResults");

    var syncAction = syncRequest.action + (isEmpty(syncRequest.id) ? "" : ", Remedy ID: " + syncRequest.id);
    var totalCount = successList.size() + warningList.size() + failedList.size();
    var successCount = successList.size();
    var warningCount = warningList.size();
    var failCount = failedList.size();


    var msg = "Synchronization Summary\n\n";

    if (syncAbortException != null)
    {
        msg += DATA_LOAD_EXCEPTION_LOG_MESSAGE + "\n";
        msg += "Exception: " + syncAbortException.toString() + "\n"
    }

    msg += "Action: " + syncAction + "\n";
    msg += "Total Number of Objects Processed: " + totalCount + "\n";
    msg += "Successful Count: " + successCount + "\n";
    msg += "Successful with Warnings Count: " + warningCount + "\n";
    msg += "Failure Count: " + failCount + "\n\n";

    log.warn(msg);

    if (log.isDebugEnabled())
    {
        if (successList.size() > 0)
        {
            log.debug("Synchronization Success Summary:");
            for (var i = 0; i < successList.size(); i++)
            {
                log.debug(successList.get(i));
            }
        }
    }

    if (warningList.size() > 0)
    {
        log.warn("Synchronization Success with Warnings Summary:");
        for (var i = 0; i < warningList.size(); i++)
        {
            log.warn(warningList.get(i));
        }
    }

    if (failedList.size() > 0)
    {
        log.warn("Synchronization Failure Summary:");
        for (var i = 0; i < failedList.size(); i++)
        {
            log.warn(failedList.get(i));
        }
    }

    if ((SEND_SYNC_SUMMARY && (syncRequest.action == ACTION_LOAD))
        || (warningList.size() > 0)
        || (failedList.size() > 0))
    {
        sendSyncSummaryMessage(syncAction, syncAbortException, totalCount, successCount, warningCount, failCount);
    }

    log.info("Exit - logSyncResults");
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * sendSyncSummaryMessage
 *
 * Sends a message to xMatters with a summary of the sync results
 *
 * @param msg String containing the message to send
 * ---------------------------------------------------------------------------------------------------------------------
 */
function sendSyncSummaryMessage(syncAction, syncAbortException, totalCount, successCount, warningCount, failCount)
{
    log.info("Enter - sendSyncSummaryMessage");

    var apxml = ServiceAPI.createAPXML();

    apxml.setMethod("ADD");
    apxml.setSubclass("action");
    apxml.setToken("recipients", XMATTERS_ADMINISTRATOR);

    apxml.setToken("action", syncAction);

    // use toString() on the numeric count values to avoid them being formatted as floating point when they get to xMatters

    apxml.setToken("total_count", totalCount.toString());
    apxml.setToken("success_count", successCount.toString());
    apxml.setToken("warning_count", warningCount.toString());
    apxml.setToken("error_count", failCount.toString());

    apxml.setToken("has_warnings", warningCount > 0 ? "true" : "false");
    apxml.setToken("has_errors", failCount > 0 ? "true" : "false");

    var notes = "";

    if (syncAbortException != null)
    {
        // Set the has_errors flag to true so subscriptions that key on this will match this notification.
        // Didn't set the error_count because that so far relates to fine-grained errors with devices, users and groups.
        apxml.setToken("has_errors", "true");
        notes += DATA_LOAD_EXCEPTION_LOG_MESSAGE + ". ";
    }

    notes += "For further information, please refer to the Integration Agent log file";

    apxml.setToken("notes", notes);

    try
    {
        ServiceAPI.sendAPXML(apxml);
        log.debug("Finished sending apxml to xMatters");
    }
    catch (e)
    {
        log.error("Caught exception processing [sync summary] Exception:" + e);
    }

    log.info("Exit - sendSyncSummaryMessage");
}

/**
 * Can be used to create any children elements that represent an list item.
 * @param nodeName name to give the node the attribute will be enclosed in
 * @param arrayList array of attributes that will be added to the collection
 * @return String containining a formatted list
 */
function createListElements(nodeName, arrayList)
{
    var listElements = "";
    arrayList.forEach(function(element)
    {
        listElements += "<" + nodeName + ">" + element.toString().trim() + "</" + nodeName + ">";
    });
    return listElements;
}

/**
 * ---------------------------------------------------------------------------------------------------------------------
 * getPassword
 *
 * Decrypts the password for the ServiceDeskUser used by the integration
 *
 * @param passwordFile file/path of the file containing the encrypted password
 * @return decrypted password or an empty string if the password cannot be decrypted
 * ---------------------------------------------------------------------------------------------------------------------
 */
function getPassword(passwordFile)
{
    try
    {
        var encryptionUtils = new EncryptionUtils();
        var file = new File(passwordFile);
        return encryptionUtils.decrypt(file);
    }
    catch (e)
    {
        return "";
    }
}

// Not used in this integration
function apia_webservice_init() {}

/**
 * This is the main method for handling APXML messages sent to the
 * Integration Service by AlarmPoint.  The messages may be responses to
 * previous submissions, or they may be requests to the Integration Service
 * that originate from AlarmPoint
 * (e.g., via an ExternalServiceMessage/ExternalServiceRequest2).
 * <p>
 * Any APXMLMessage object that this method returns will be sent to AlarmPoint
 * either via the Integration Service's outbound queues or directly as a
 * Web Service response, depending on the mechanism used to submit the APXML.
 */
function apia_response(apxml)
{
    var method = apxml.getMethod();

    if (APXMLMessage.OK_METHOD.equalsIgnoreCase(method))
    {
        return handleOK(apxml);
    }
    else if (APXMLMessage.ERROR_METHOD.equalsIgnoreCase(method))
    {
        return handleERROR(apxml);
    }
    else if (APXMLMessage.SEND_METHOD.equalsIgnoreCase(method))
    {
        return handleSend(apxml);
    }
    else if (APXMLMessage.REQUEST_METHOD.equalsIgnoreCase(method))
    {
        return handleRequest(apxml);
    }
    else
    {
        throw new IllegalArgumentException("Unrecognized APXML method: " + method);
    }
}

/**
 * OK responses are returned by AlarmPoint for each APXML message that it accepts
 * from an Integration Service.
 * <p>
 * Log the acknowledgement using the ping Integration Service's logger.
 */
function handleOK(apxml)
{
    var id = apxml.getTransactionId();
    ServiceAPI.getLogger().info("Submission of transaction " + id + "succeeded.");
}

/**
 * ERROR responses are returned by AlarmPoint for each APXML message that it
 * is unable to accept from an Integration Service.
 * <p>
 * Log the error using the ping Integration Service's logger.
 */
function handleERROR(apxml)
{
    var id = apxml.getTransactionId();
    var code = apxml.getValue(APXMLMessage.ERROR_CODE);
    ServiceAPI.getLogger().error("Submission of transaction " + id + "failed with error code " + code);
}

/**
 * Requests are forwarded to the Send handler.
 * <p>
 * Redirects to the handleSend as messages may come from ESM or ESR to the IA.
 */
function handleRequest(apxml)
{
    log.info("handleRequest received apxml, forwarding to handleSend");
    return handleSend(apxml);
}

/**
 * The main routine for handling Requests AlarmPoint
 * <p>
 * Throw an exception that will be logged by the Integration Agent.
 */
function handleSend(apxml)
{
    var id = apxml.getTransactionId();
    log.info("Response from xMatters for transaction " + id + " received.");
}
