importPackage(java.lang);
importPackage(java.util);
importPackage(java.io);
importClass(Packages.com.xmatters.ca.servicedesk.util.ParseCAFile);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLMessage);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLToken);
importClass(Packages.com.thoughtworks.xstream.XStream);
importClass(Packages.com.thoughtworks.xstream.converters.reflection.PureJavaReflectionProvider);
importClass(Packages.com.alarmpoint.integrationagent.security.EncryptionUtils);
importClass(Packages.com.alarmpoint.integrationagent.util.EventDeduplicator);

var xStream = new XStream(new PureJavaReflectionProvider())

load("integrationservices/caservicedesk/caservicedesk/baseclass.js");
load("integrationservices/caservicedesk/caservicedesk/logger.js");
load("integrationservices/caservicedesk/caservicedesk/configuration.js");
load("integrationservices/caservicedesk/caservicedesk/wsutil.js");
load("integrationservices/caservicedesk/caservicedesk/caservicedeskws.js");
load("integrationservices/caservicedesk/caservicedesk/transitions.js");
load("integrationservices/caservicedesk/caservicedesk/xmattersws.js");
load("integrationservices/caservicedesk/caservicedesk/xMattersEventWS.js");

// Mobile Gateway includes
load("integrationservices/caservicedesk/caservicedesk/mobilegateway.js");

var exceptionString = "JavaException: ";

var notePrefix = "[xMatters] - "; // The Prefix which is added to the note comments for AlarmPoint e.g. "[xMatters] " 

var TOKEN_NAME = "apiaResult";
var TOKEN_NAME_EXCEPTION = "apiaResultException";

var SERVICE_DESK_PASSWORD = getPassword(SERVICE_DESK_PASSWORD_FILE);

/**
 * This is the main method for handling APXML messages sent to the 
 * Integration Service by APClient.  APClient sends the location of the CA Service Desk notification file.
 * This notification file should be created as part of the xMatters notification method.  The file is then
 * parsed for parameters which will be sent to AlarmPoint as event tokens.
 * <p>
 * For example, the key NX_NTF_MESSAGE_HTML is found in the notification file and is sent to AlarmPoint
 * as NOTIFICATION_MESSAGE_HTML_0 , NOTIFICATION_MESSAGE_HTML_1.  It is broken up as each event token is
 * limited to 4000k.  Once recieved by AlarmPoint these tokens are re-assembled and sent as an html notification
 * to the targeted recipient.
 * <p>
 * Any APXMLMessage object that this method returns will be sent to AlarmPoint
 * via the Integration Service's outbound queues.
 */
function apia_input(apxml) 
{
  // Parse file
  var notificationFile = apxml.getValue("notification_file");
  if(notificationFile == "sync")
  {
    throw "Data Load should now be used from the casddataload integration service.  Targeting this service is now an error."

    var caServiceDeskWS = new CAServiceDeskWS(SERVICE_DESK_URL);
    caServiceDeskWS.login(SERVICE_DESK_USER, SERVICE_DESK_PASSWORD);
    
    try
    {
      // Reset the success/warning/failure collections 
      failedList = new ArrayList();
      successList = new ArrayList();
      warningList = new ArrayList();
      reprocessedList = new ArrayList();
      
      syncUsersAndDevices(caServiceDeskWS);
      syncGroups(caServiceDeskWS);
      logSyncResults(caServiceDeskWS);
    }
    finally
    {
      caServiceDeskWS.logout();
    }
  }
  else
  {
    return handleNotifications(apxml, notificationFile);
  } 
}


/**
 * Handle the notification from CA Service Desk
 * @param apxml notification content
 * @param notificationFile name of the notification file to be parsed
 * @return enriched APXML to send to xMatters
 */
function handleNotifications(apxml, notificationFile)
{
  var caServiceDeskWS = new CAServiceDeskWS(SERVICE_DESK_URL);

  log.debug("Notification file: " + notificationFile);
  var fileParser = new ParseCAFile(notificationFile);
  var parsedFile = fileParser.parse();
  // markProcessed() will attempt to move the parsed notification file to a subdirectory called xmatters.
  // This is bound by file system access so if you don't want this feature just comment out fileParser.markProcessed().
  fileParser.setCopyAfterProcessing(false);
  fileParser.markProcessed();
  //fileParser.markProcessed('<custom directory>'); // If you want a different directory, use full qualified path.
  //fileParser.markProcessed('<custom directory>', '<custom filename>'); // If you want to control directory and file name.
  
  // Add object section to APXML
  var parsedFileKeys = (parsedFile.keySet()).toArray();
  var objectSectionKey;
  for(var i=0, len=parsedFileKeys.length; i<len; ++i)
  {
    if(parsedFileKeys[i] != "cnt" && parsedFileKeys[i] != "notification")
    {
      objectSectionKey = parsedFileKeys[i];
    }
  }
  var objectSection = parsedFile.get(objectSectionKey);

  if ("chg".equalsIgnoreCase(objectSectionKey))
  {
    apxml.setToken("NX_NTF_REF_NUM", objectSection.get("NX_NTF_CHG_REF_NUM"));
    log.debug("NX_NTF_REF_NUM: " + objectSection.get("NX_NTF_REF_NUM"));
  }
  else
  {
    apxml.setToken("NX_NTF_REF_NUM", objectSection.get("NX_NTF_REF_NUM"));
    log.debug("NX_NTF_REF_NUM: " + objectSection.get("NX_NTF_REF_NUM"));
  }
  apxml.setToken("NX_NTF_PERSISTENT_ID", objectSection.get("NX_NTF_PERSISTENT_ID"));
  apxml.setToken("incident_id", objectSection.get("NX_NTF_PERSISTENT_ID"));
  apxml.setToken(APXMLMessage.APIA_PROCESS_GROUP, objectSection.get("NX_NTF_PERSISTENT_ID"));
  log.debug("NX_NTF_TYPE_SYM: " + objectSection.get("NX_NTF_TYPE.SYM"));
  apxml.setToken("NX_NTF_TYPE_SYM", objectSection.get("NX_NTF_TYPE.SYM"));
  log.debug("NX_NTF_STATUS_SYM: " + objectSection.get("NX_NTF_STATUS.SYM"));
  apxml.setToken("NX_NTF_STATUS_SYM", objectSection.get("NX_NTF_STATUS.SYM"));
  log.debug("NX_NTF_PRIORITY_SYM: " + objectSection.get("NX_NTF_PRIORITY.SYM"));
  apxml.setToken("NX_NTF_PRIORITY_SYM", objectSection.get("NX_NTF_PRIORITY.SYM"));
  log.debug("NX_NTF_CUSTOMER_COMBO_NAME: " + objectSection.get("NX_NTF_CUSTOMER.COMBO_NAME"));
  apxml.setToken("NX_NTF_CUSTOMER_COMBO_NAME", objectSection.get("NX_NTF_CUSTOMER.COMBO_NAME"));
  log.debug("NX_NTF_ASSIGNEE_COMBO_NAME: " + objectSection.get("NX_NTF_ASSIGNEE.COMBO_NAME"));
  apxml.setToken("NX_NTF_ASSIGNEE_COMBO_NAME", objectSection.get("NX_NTF_ASSIGNEE.COMBO_NAME"));
  log.debug("NX_NTF_DESCRIPTION: " + objectSection.get("NX_NTF_DESCRIPTION"));
  apxml.setToken("NX_NTF_DESCRIPTION", objectSection.get("NX_NTF_DESCRIPTION"));
  log.debug("TICKET_SUMMARY: " + objectSection.get("NX_NTF_SUMMARY"));
  apxml.setToken("TICKET_SUMMARY", objectSection.get("NX_NTF_SUMMARY"));
  
  // Add notification section to APXML
  var notificationSection = parsedFile.get("notification");
  apxml.setToken("NOTIFICATION_SUMMARY", notificationSection.get("NX_NTF_SUMMARY"));
  log.debug("NOTIFICATION_MESSAGE: " + notificationSection.get("NX_NTF_MESSAGE"));
  apxml.setToken("NOTIFICATION_MESSAGE", notificationSection.get("NX_NTF_MESSAGE"));

  var notification_userid = notificationSection.get("NX_NTF_USERID");  
  log.debug("NOTIFICATION_USERID: " + notification_userid);
  apxml.setToken("NOTIFICATION_USERID", notification_userid);
  log.debug("NOTIFICATION_COMBO_NAME: " + notificationSection.get("NX_NTF_COMBO_NAME"));
  apxml.setToken("NOTIFICATION_COMBO_NAME", notificationSection.get("NX_NTF_COMBO_NAME"));
  var nx_ntf_urgency = notificationSection.get("NX_NTF_URGENCY");
  // nx_ntf_urgency is represented in the CA SDM notification file as an integer.
  // There is a set of key value pairs in the notification file which seem to indicate that
  // 1 maps to a low notification and 4 maps to an emergency notification.
  // This association is the assumption made to create the if block below.
  // Symbols from notification file:
  // NX_NTF_NOTIFY_URGENCY1.SYM=Low
  // NX_NTF_NOTIFY_URGENCY2.SYM=Normal
  // NX_NTF_NOTIFY_URGENCY3.SYM=High
  // NX_NTF_NOTIFY_URGENCY4.SYM=Emergency
  if(nx_ntf_urgency > 2)
  {
    apxml.setToken("apia_priority", "high");
    log.debug("APIA_PRIORITY: " + notificationSection.get("NX_NTF_URGENCY"));
  }
    
  // Add cnt section to APXML
  var cntSection = parsedFile.get("cnt");
  apxml.setToken("CNT_PERSISTENT_ID", cntSection.get("NX_NTF_PERSISTENT_ID"));
  log.debug("CNT_PERSISTENT_ID: " + cntSection.get("NX_NTF_PERSISTENT_ID"));
  apxml.setToken("CNT_NOTIFY_URGENCY_SYM", cntSection.get("NX_NTF_NOTIFY_URGENCY" + nx_ntf_urgency + ".SYM"));
  log.debug("CNT_NOTIFY_URGENCY_SYM: " + cntSection.get("NX_NTF_NOTIFY_URGENCY" + nx_ntf_urgency + ".SYM"));

  // Subscription information
  log.debug("REQUESTOR_NAME: " + objectSection.get("NX_NTF_REQUESTOR.COMBO_NAME"));
  apxml.setToken("REQUESTOR_NAME", objectSection.get("NX_NTF_REQUESTOR.COMBO_NAME"));
  log.debug("ASSIGNEE_NAME: " + notificationSection.get("NX_NTF_ASSIGNEE"));
  apxml.setToken("ASSIGNEE_NAME", notificationSection.get("NX_NTF_ASSIGNEE"));
  log.debug("PRIORITY: " + objectSection.get("NX_NTF_PRIORITY.SYM"));
  apxml.setToken("PRIORITY", objectSection.get("NX_NTF_PRIORITY.SYM"));
  log.debug("STATUS: " + objectSection.get("NX_NTF_STATUS.SYM"));
  apxml.setToken("STATUS", objectSection.get("NX_NTF_STATUS.SYM"));
  log.debug("URGENCY: " + objectSection.get("NX_NTF_URGENCY.SYM"));
  apxml.setToken("URGENCY", objectSection.get("NX_NTF_URGENCY.SYM"));
  log.debug("IMPACT: " + objectSection.get("NX_NTF_IMPACT.SYM"));
  apxml.setToken("IMPACT", objectSection.get("NX_NTF_IMPACT.SYM"));
  log.debug("SEVERITY: " + objectSection.get("NX_NTF_SEVERITY.SYM"));
  apxml.setToken("SEVERITY", objectSection.get("NX_NTF_SEVERITY.SYM"));

  if(notification_userid != null && notification_userid != "")
  {
  	log.debug("Setting recipients to value of NX_NTF_USERID: " + notification_userid);
  	apxml.setToken("recipients", notification_userid);
  }
  else
  {
    log.debug("Setting recipients to value of NX_NTF_COMBO_NAME: " + notificationSection.get("NX_NTF_COMBO_NAME"));
    apxml.setToken("recipients", notificationSection.get("NX_NTF_COMBO_NAME"));
  }
  
  // Break up the HTML String into 4000 char chunks so that it can be processed by the APS.
  var htmlString = notificationSection.get("NX_NTF_MESSAGE_HTML");
  if (htmlString != null)
  {
    // Java String
    log.debug("HTML String length: " + htmlString.length());
    apxml.setToken("CA_HTML_NOTIFICATION_LENGTH", htmlString.length());
    htmlString = htmlString.replaceAll("[\\r\\n]", "");
    var chunkedString = htmlString.match(RegExp('.{1,4000}', 'g'));

    // JavaScript String
    log.debug("Size of chunk array: " + chunkedString.length);
    apxml.setToken("CA_HTML_NOTIFICATION_CHUNKS", chunkedString.length);
    for(var i=0, len=chunkedString.length; i<len; ++i)
    {
      log.debug("HTML chunk "+ i + ": " + chunkedString[i]);
      apxml.setToken("XM_NOTIFICATION_MESSAGE_HTML_"+i, chunkedString[i]);
    }
  }
  
  // Add the valid response choices to the APXML
  var ticketFactory = objectSectionKey;
  caServiceDeskWS.login(SERVICE_DESK_USER, SERVICE_DESK_PASSWORD);
  if ("cr".equalsIgnoreCase(ticketFactory))
  {
    // This is Problem/Incident/Request...retrieve the ticket factory information for the transitions
    ticketFactory = caServiceDeskWS.getTicketFactory(objectSection.get("NX_NTF_TYPE.SYM"))
  }

  var responseList
  if (USE_DYNAMIC_TRANSITIONS)
  {
    responseList = caServiceDeskWS.getResponseOptions(objectSection.get("NX_NTF_PERSISTENT_ID"), ticketFactory);
  }
  else
  {
    responseList = getStaticResponseOptions(ticketFactory, objectSection.get("NX_NTF_STATUS.SYM"));
  }
  apxml.setToken("TICKET_FACTORY", ticketFactory);
  apxml.setToken("RESPONSE_OPTIONS", responseList.toString());
  apxml.setToken("RESPONSE_OPTIONS-list-item-delimiter", ",");

  // Logout of the webservice session
  caServiceDeskWS.logout();


  // Send enriched APXML
  var recentlyReceived = EventDeduplicator.getInstance().recentOccurrenceCount(apxml, DEDUPLICATOR_FILTER); 
  if (recentlyReceived <= 1) 
  {
    // Before sending the new event to xMatters, terminate any existing events for the given incident_id  
    handleDel(objectSection.get("NX_NTF_PERSISTENT_ID"),apxml.getValue("apia_priority"));
    return apxml; 
  } 
  else 
  { 
    log.warn("***Deduplicator Suppressed Notification ***: An event with tokens " + convertTokensToString(apxml) + " has been injected into the " + 
             apxml.getValue("agent_client_id") + " event domain " + recentlyReceived + " times within the configured suppression period.  It has been suppressed." ); 
  }
}

// Not used in this integration, but the 5.0 IA wants this function to exist
function apia_webservice_init() {}

/**
 * This is the main method for handling APXML messages sent to the
 * Integration Service by AlarmPoint.  The messages may be responses to 
 * previous submissions, or they may be requests to the Integration Service
 * that originate from AlarmPoint 
 * (e.g., via an ExternalServiceMessage/ExternalServiceRequest2).
 * <p>
 * Any APXMLMessage object that this method returns will be sent to AlarmPoint
 * either via the Integration Service's outbound queues or directly as a 
 * Web Service response, depending on the mechanism used to submit the APXML.
 */
function apia_response(apxml) 
{
  var method = apxml.getMethod();
  
  if (APXMLMessage.OK_METHOD.equalsIgnoreCase(method))
  {
    return handleOK(apxml);
  } 
  else if (APXMLMessage.ERROR_METHOD.equalsIgnoreCase(method))
  {
    return handleERROR(apxml);
  } 
  else if (APXMLMessage.SEND_METHOD.equalsIgnoreCase(method))
  {
    return handleSend(apxml);
  }
  else if (APXMLMessage.REQUEST_METHOD.equalsIgnoreCase(method))
  {
    return handleRequest(apxml);
  }
  else
  {
    throw new IllegalArgumentException("Unrecognized APXML method: " + method);
  }
}

/**
 * OK responses are returned by AlarmPoint for each APXML message that it accepts
 * from an Integration Service.
 * <p>
 * Log the acknowledgement using the ping Integration Service's logger.
 */
function handleOK(apxml)
{
  var id = apxml.getTransactionId();
  ServiceAPI.getLogger().info("Submission of transaction " + id + "succeeded.");
}

/**
 * ERROR responses are returned by AlarmPoint for each APXML message that it
 * is unable to accept from an Integration Service.
 * <p>
 * Log the error using the ping Integration Service's logger.
 */
function handleERROR(apxml) 
{
  var id = apxml.getTransactionId();
  var code = apxml.getValue(APXMLMessage.ERROR_CODE); 
  ServiceAPI.getLogger().error("Submission of transaction " + id + "failed with error code " + code);
}

/**
 * Requests are forwarded to the Send handler.
 * <p>
 * Redirects to the handleSend as messages may come from ESM or ESR to the IA.
 */
function handleRequest(apxml)
{
  return handleSend(apxml);
}

/**
 * The main routine for handling del injections from the management system to AlarmPoint
 * As the CA SDM notifications are not based on event, a DEL event will be sent ahead of
 * every add event to terminate any existing events for the ticket.
 * @param incidentId id of the incident event that will be terminated
 */  
function handleDel(incidentId, priority)
{
  var xmEventWS = new xMattersEventWS();
  // Check if the notification exists in xMatters
  if (xmEventWS.queryIncident(incidentId))
  {
    var apxml = ServiceAPI.createAPXML();
    apxml.setMethod("DEL");
    apxml.setSubclass("action");
    apxml.setToken("apia_priority", priority);
    apxml.setToken(APXMLMessage.APIA_PROCESS_GROUP, incidentId);
    apxml.setToken("incident_id", incidentId);

    log.debug("handleDel apxml: " + apxml);
    
    var response = initializeResponse(apxml);
    try
    {
      ServiceAPI.sendAPXML(apxml);      
      response.setToken(TOKEN_NAME, "OK", APXMLToken.Type.STRING);
      log.debug("Finished sending apxml to AlarmPoint");
    }
    catch (e)
    {
      log.error("Caught exception processing [DEL]. Exception:" + e);
      response.setToken(TOKEN_NAME_EXCEPTION, e.toString().replace(exceptionString, ""), APXMLToken.Type.STRING);
    }

    log.debug(response);
  }
}


/**
 * The main routine for handling Requests AlarmPoint
 * <p>
 * Throw an exception that will be logged by the Integration Agent.
 */
function handleSend(apxml)
{
  log.debug( "handleSend received apxml: " + apxml ); 

  // handle annotations coming from callout APS script 
  // they are formatted differently 
  // the presence of the message token means this message 
  // is a callout annotation.
  var message = apxml.getValue("message");
  var incidentId = apxml.getToken("incident_id").getValue();
  
  if ( message != null && incidentId != null)
  {
    return handleCalloutAnnotate( apxml );
  }
  else
  {
    return handleApsResponses( apxml );
  }
}

/**
 * Used to update CA Service Desk Incidents using web services
 * <p>
 * The main routine for handling user responses to xMatters notifications.
 * <p>
 * Throw an exception that will be logged by the Integration Agent.
 */
function handleApsResponses(apxml) 
{
  var caServiceDeskWS = new CAServiceDeskWS(SERVICE_DESK_URL);
  var response = initializeResponse(apxml);

  var incidentId = apxml.getToken("incident_id").getValue();
  var requestText = "";
  var messageText = notePrefix + "Updated Task";
  if(apxml.getToken("request_text") != null)
  {
    requestText = apxml.getToken("request_text").getValue();
  }
  if(apxml.getToken("message_text") != null && !"".equals(apxml.getToken("message_text")))
  {
    messageText = notePrefix + apxml.getToken("message_text").getValue();
  }

  // deserialize the additional tokens
  var additionalTokens = getAdditionalTokens(apxml);
  log.debug( "additionalTokens keys: " + additionalTokens);

  var objectHandle = processAdditionalTokens(additionalTokens, "object_handle", false);     
  var target = processAdditionalTokens(additionalTokens, "target", false);     

  log.debug("About to start handling " + requestText);

  if (objectHandle != null && !"".equals(objectHandle))
  {
    try
    {
    // Start a webservice session
      if (!MANAGED_LOGIN)
      {
        caServiceDeskWS.login(SERVICE_DESK_USER, SERVICE_DESK_PASSWORD);
      }
      else
      {
        caServiceDeskWS.login(target, null);
      }
  
      if ( !requestText.equalsIgnoreCase( "Annotate" ) && !requestText.equalsIgnoreCase( "IGN" ) )
      {
        caServiceDeskWS.updateTicket(incidentId, requestText, additionalTokens, objectHandle);
      }
      caServiceDeskWS.annotate(objectHandle, caServiceDeskWS.getHandleForUserId((target != null ? target : SERVICE_DESK_USER)), messageText);
      response.setToken(TOKEN_NAME, "OK", APXMLToken.Type.STRING);
    }
    catch (e)
    {
      log.error("Caught exception processing [" + requestText + "] Exception:" + e);
      response.setToken(TOKEN_NAME_EXCEPTION, e.toString().replace(exceptionString, ""), APXMLToken.Type.STRING);
    }
    finally
    {
      // Logout of the webservice session
      caServiceDeskWS.logout();
    }
  }
  
  log.debug("Finished call to management system for requestText: [" + requestText + "]");
  log.debug("Response: " + response);

  return response;
}

/**
 * handles annotations sent from the Callout APS script
 */
function handleCalloutAnnotate( apxml )
{
  var caServiceDeskWS = new CAServiceDeskWS(SERVICE_DESK_URL);
  var messageText = apxml.getValue("message");

  var additionalTokens = getAdditionalTokens(apxml);
  log.debug( "additionalTokens keys: " + additionalTokens);
  var incidentId = processAdditionalTokens(additionalTokens, "incident_id", true); 
  var objectHandle = processAdditionalTokens(additionalTokens, "object_handle", true); 
    
  log.debug( "About to call to management system for callout annotate for [" + incidentId + "]" );

  try 
  {
    caServiceDeskWS.login(SERVICE_DESK_USER, SERVICE_DESK_PASSWORD);
    caServiceDeskWS.annotate(objectHandle, caServiceDeskWS.getHandleForUserId(SERVICE_DESK_USER), messageText);
  }
  catch (e)
  {
    log.error("Caught exception processing annotation for [" + incidentId + "] Exception:" + e);
  }
  finally
  {
    caServiceDeskWS.logout();
  }
  
  log.debug("Finished call to management system for callout annotate");
}


/**
 * Retrieves the additional tokens from the APXML
 * @param apxml XML message to process
 * @return additional tokens
 * @throws if the additional tokens are missing from the XML.
 */
function getAdditionalTokens(apxml)
{
  if (apxml.getValue("additionalTokens") != null && !"".equalsIgnoreCase(apxml.getValue("additionalTokens"))) 
  {
    return xStream.fromXML(apxml.getValue("additionalTokens")); // LinkedHashMap
  }
  throw new Exception("additionalTokens is missing from apxml"); 
}

/**
 * Retrieves the additional tokens from the APXML and throws an exception if a mandatory token is missing
 * @param additionalTokens collection of tokens to process
 * @param tokenName name of the token to retrieve
 * @param mandatory true if token is required, false otherwise
 * @return the token value
 */
function processAdditionalTokens(additionalTokens, tokenName, mandatory)
{
	var token = additionalTokens.get(tokenName);
	if (mandatory && token == null)
	{
		throw new Exception("additionalTokens in apxml is missing " + tokenName + " which is mandatory");
	}
	return token;
}

/**
 * Initializes the response to be returned to the device
 * @param apxml APXML containing the request information
 * @return initialized instance of the response APXML
 */
function initializeResponse(apxml)
{
  var response = ServiceAPI.createAPXML();
  response.setMethod(APXMLMessage.RESPONSE_METHOD);
  response.setSubclass(APXMLMessage.RESPONSE_SUBCLASS);
  
  if (apxml.hasToken(APXMLMessage.REQUEST_ID))
  {
    response.setToken(APXMLMessage.REQUEST_ID, apxml.getValue(APXMLMessage.REQUEST_ID), APXMLToken.Type.STRING);
  }
  response.setToken(APXMLMessage.DESTINATION, apxml.getValue(APXMLMessage.ORIGINATOR), APXMLToken.Type.STRING);
  if (apxml.hasToken(APXMLMessage.APIA_PRIORITY)) 
  {
    response.setToken(APXMLMessage.APIA_PRIORITY, apxml.getValue(APXMLMessage.APIA_PRIORITY));
  }
  else
  {
    response.setToken(APXMLMessage.APIA_PRIORITY, APXMLMessage.APIA_PRIORITY_HIGH);
  }
  return response;
}

//////////////////////////////////////////////////////////////////////////////////////////////////
//                                    DATA SYNC HELPER METHODS                                  //
//////////////////////////////////////////////////////////////////////////////////////////////////

/**
 * Creates a where clause for the web service call to filter the user/groups retrieved
 * @param nodeName name of the parent node containing the ids that are required
 * @param type user or group (to retrieve the correct data from the list)
 * @param fieldName attribute name used to identify the data
 */
function getSyncFilter(nodeName, type, fieldName)
{
  var nodes = syncList.child(nodeName);
  var whereClause = new StringBuffer("(");
  var exclude = SYNC_ACTION.equalsIgnoreCase("exclude");
  var operator = exclude ? "!=" : "=";
  var concat = exclude ? " AND " : " OR ";

  // Iterate the users/groups in the selected node and build up the where clause
  for (var i=0; i<nodes.child(type.toLowerCase()).length(); i++)
  {
    var name = nodes.child(type.toLowerCase())[i];
    whereClause.append(((i>0) ? concat : ""));
    whereClause.append(fieldName + " " + operator + " '" + name.toString().replace("'", "''") + "'");
    log.debug(type + ": " + name + " will " + (exclude ? "NOT" : "") + " be synchronized as " + name + " is on the " + (exclude ? "EXCLUDE" : "INCLUDE") + " list");  
  }
  whereClause.append(")");
  return whereClause.toString();
}

/**
 * Synchronize the users and devices with xMatters
 * @param caServiceDeskWS instance of the CA Service Desk Web Service object
 */
function syncUsersAndDevices(caServiceDeskWS) 
{
  log.debug("*** Starting User and Device Sync ***");
  var groupId = caServiceDeskWS.getContactTypeId("Group");
  var objectType = "cnt";
  var whereClause = "(type is null OR type != " + groupId + ") AND " + getSyncFilter("users", "User", "userid");
  
  var doQueryReturn = caServiceDeskWS.doQuery(objectType, whereClause);
  var listLength = doQueryReturn.listLength;
  var listHandle = doQueryReturn.listHandle;

  var numberOfChunks = Math.ceil(listLength / CHUNK_SIZE);
  log.debug("Number of chunks: " + numberOfChunks);
  
  for (var i=0; i<numberOfChunks; i++)
  {
    var startIndex = i * CHUNK_SIZE;
    log.debug("startIndex: " + startIndex);
    var udsObjectListArray = caServiceDeskWS.processListChunk(objectType, whereClause, caServiceDeskWS.ATTRIBUTES_CNT_OBJECT, startIndex, CHUNK_SIZE, false, listHandle, listLength);
    caServiceDeskWS.processUDSObjectList(udsObjectListArray, caServiceDeskWS.createSyncUsers);  
  }
}

/**
 * Synchronize the groups with xMatters
 * @param caServiceDeskWS instance of the CA Service Desk Web Service object
 */
function syncGroups(caServiceDeskWS) 
{
  log.debug("*** Starting Group Sync ***");
  var groupId = caServiceDeskWS.getContactTypeId("Group");
  var groupSync = new xMattersGroupSync(syncList);
  var dataSyncHelper = new DataSyncHelper(syncList, "users");

  // returns a ArrayList of Groups
  var objectType = "cnt";
  var whereClause = "type = " + groupId + " AND " + getSyncFilter("groups", "Group", "last_name");

  var doQueryReturn = caServiceDeskWS.doQuery(objectType, whereClause);
  var listLength = doQueryReturn.listLength;
  var listHandle = doQueryReturn.listHandle;

  var numberOfChunks = Math.ceil(listLength / CHUNK_SIZE);
  log.debug("Number of chunks: " + numberOfChunks);
  
  for (var j=0; j<numberOfChunks; j++)
  {
    var startIndex = j * CHUNK_SIZE;
    log.debug("startIndex: " + startIndex);

    var udsObjectListArray = caServiceDeskWS.processListChunk(objectType, whereClause, caServiceDeskWS.ATTRIBUTES_CNT_OBJECT, startIndex, CHUNK_SIZE, false, listHandle, listLength);
    var groups = caServiceDeskWS.processUDSObjectList(udsObjectListArray, caServiceDeskWS.createSyncGroups);  

    // check to see if groups are returnd
    if (groups != null && groups.size() > 0) 
    {
      for (var i=0; i<groups.size();i++)
      {
        var group = groups.get(i);      
  
        // returns an ArrayList with the member ID
        log.debug("Looking for group members of group: " + group.name);
        var memberIdArray = caServiceDeskWS.processUDSObjectList(caServiceDeskWS.processListValues("grpmem", "group = U'" + group.id + "'", caServiceDeskWS.ATTRIBUTES_GRPMEM_OBJECT), caServiceDeskWS.processGroupMembers);
        if (memberIdArray != null && memberIdArray.size() > 0)
        {
          for (var x=0; x < memberIdArray.size(); x++ )
          {
            // returns an ArrayList with the member
            var member = caServiceDeskWS.processUDSObjectList(caServiceDeskWS.processListValues("cnt", "id = U'" + memberIdArray.get(x) + "'", caServiceDeskWS.ATTRIBUTES_CNT_OBJECT), caServiceDeskWS.retrieveGroupMemberRecord);
  
            if (member != null && member.size() > 0) 
            {
              var memberName = member.get(0).targetName;
              if (dataSyncHelper.checkIncludeExcludeList("User", memberName))
              {
                log.debug("Adding Group Member: " + memberName);
                group.team.members.put(memberName, member.get(0));
              }
            }
          }
        }
        groupSync.syncGroup(group);
      }
    }
  }  
}

/**
 * Logs the results from the synchronization process
 * @param caServiceDeskWS instance of the CA Service Desk Web Service object
 */
function logSyncResults(caServiceDeskWS)
{
  var objectCount = successList.size() + warningList.size() + failedList.size();
  var msg = "Synchronization Summary\n\n";
  msg += "Total Number of Objects Processed: " + objectCount + "\n";
  msg += "Successful Count: " + successList.size() + "\n";
  msg += "Successful with Warnings Count: " + warningList.size() + "\n";
  msg += "Failure Count: " + failedList.size() + "\n\n";
  caServiceDeskWS.log.warn(msg);
  msg += "For further information, please refer to the default logs in /integrationagent/log/AlarmPoint.txt.";

  if (caServiceDeskWS.log.isDebugEnabled())
  {
    if (successList.size() > 0)
    {
      caServiceDeskWS.log.debug("Synchronization Success Summary:");
      for (var i=0; i<successList.size(); i++)
      {
        caServiceDeskWS.log.debug(successList.get(i));
      }
    }
  }

  if (warningList.size() > 0)
  {
    caServiceDeskWS.log.warn("Synchronization Success with Warnings Summary:");
    for (var i=0; i<warningList.size(); i++)
    {
     caServiceDeskWS.log.warn(warningList.get(i));
    }
  }

  if (failedList.size() > 0) 
  {
    caServiceDeskWS.log.warn("Synchronization Failure Summary:");
    for (var i=0; i<failedList.size(); i++)
    {
      caServiceDeskWS.log.warn(failedList.get(i));
    }
  }
  
  if (SEND_SYNC_SUMMARY)
  {
    sendQuickMessage(msg);
  }
}

/**
 * Sends a quick message to xMatters
 * @param msg String containing the message to send
 */
function sendQuickMessage(msg)
{
  var apxml = ServiceAPI.createAPXML();
  apxml.setMethod("ADD");
  apxml.setSubclass("action");
  apxml.setToken("message", msg);
  apxml.setToken("quickmessage", "true");
  apxml.setToken("recipients", XMATTERS_ADMINISTRATOR);
  apxml.setToken("subject", "CA Service Desk Data Synchronization Summary");

  var response = initializeResponse(apxml);
  try
  {
    ServiceAPI.sendAPXML(apxml);      
    response.setToken(TOKEN_NAME, "OK", APXMLToken.Type.STRING);
    log.debug("Finished sending apxml to AlarmPoint");
  }
  catch (e)
  {
    log.error("Caught exception processing [quickMessage] Exception:" + e);
    response.setToken(TOKEN_NAME_EXCEPTION, e.toString().replace(exceptionString, ""), APXMLToken.Type.STRING);
  }
  log.debug(response);
}

/**
 * Decrypts the password for the ServiceDeskUser used by the integration
 * @param passwordFile file/path of the file containing the encrypted password
 * @return decrypted password or an empty string if the password cannot be decrypted
 */
function getPassword(passwordFile)
{
  try
  {
    var encryptionUtils = new EncryptionUtils();
    var file = new File(passwordFile);
    return encryptionUtils.decrypt(file);
  }
  catch (e)
  {
    return "";
  }
}

/** 
 * Convert all tokens for this message into a parenthesis-grouped set of name-value-pairs. 
 * @param apxml tokens to be converted into a string
 * @return string representation of input tokens 
 */ 
function convertTokensToString(apxml) 
{ 
  var string = ""; 
  var tokenIterator = apxml.getTokens(); 
  while (tokenIterator.hasNext()) 
  { 
    var token = tokenIterator.next(); 
    string += "(" + token.getKey() + ", " + token.getValue() + ")"; 
  } 
  return string; 
}
