importClass(Packages.com.alarmpoint.integrationagent.http.HttpClientWrapper);
importClass(Packages.com.alarmpoint.integrationagent.http.DefaultX509TrustManager);
importClass(Packages.com.alarmpoint.integrationagent.soap.SOAPMessage);
importClass(Packages.com.alarmpoint.integrationagent.soap.SOAPRequest);
importClass(Packages.com.alarmpoint.integrationagent.soap.exception.SOAPRequestException);
importClass(Packages.javax.net.ssl.SSLContext);
importClass(Packages.javax.net.ssl.TrustManager);
importClass(Packages.org.apache.commons.lang.StringEscapeUtils);
importClass(Packages.org.apache.http.conn.ssl.SSLSocketFactory);

// Add functionality to the JavaScript String prototype
String.prototype.trim = function() {
	return this.replace(/^\s+|\s+$/g,"");
}
String.prototype.ltrim = function() {
	return this.replace(/^\s+/,"");
}
String.prototype.rtrim = function() {
	return this.replace(/\s+$/,"");
}
String.prototype.endsWith = function(pattern) {
    var d = this.length - pattern.length;
    return d >= 0 && this.lastIndexOf(pattern) === d;
}
String.prototype.startsWith = function(str){
    return (this.indexOf(str) === 0);
}

var WSUtil = BaseClass.extend({

  LOG_SOURCE: "wsutil.js: ",
  // 
  // Use SSL but trust any certificate (including self-signed ones).
  // This SHOULD NOT be used for productive systems due to security reasons, 
  // unless it is a conscious decision and you are perfectly aware of security 
  // implications of accepting self-signed certificates  
  ACCEPT_ANY_CERTIFICATE: true,
  
  // File/Path for the Java Keystore containing the trusted certificate chain
  KEY_STORE: "jre/lib/security/cacerts",
  
  // Password for the Java Keystore containing the trusted certificate chain
  KEY_STORE_PASSWORD: "changeit",
  
  // SSL Socket Factory used for SSL communication
  SSL_SOCKET_FACTORY: null,

  /**
   * Constructor
   */
  init: function()
  {
    this.log = new Logger(this.LOG_SOURCE);
  },

  /**
   * Sends the request to the SOAP endpoint and returns the response
   * @param url EndPoint for the webservice request
   * @param msg SOAPMessage instance containing the SOAPEnvelope to post
   * @param headers array of SOAP Headers to be added to the SOAP Request
   * @return SOAP Response Body
   */
  sendReceive: function(url, msg, headers)
  {
    // These variables are used by E4X to access nodes prefixed with a namespace. These namespaces are defined
    // in the SOAP Response that is parsed by E4X
    var soap = new Namespace("http://schemas.xmlsoap.org/soap/envelope/");

    var request;
    if (url.toLowerCase().startsWith("https"))
    {
      if (this.ACCEPT_ANY_CERTIFICATE)
      {
        request = new SOAPRequest(url, msg.getAction(), this.ACCEPT_ANY_CERTIFICATE);
        this.log.debug("SSL enabled to accept ANY certificate");
      }
      else
      {
        request = new SOAPRequest(url, msg.getAction(), this.getSSLSocketFactory());
        this.log.debug("SSL enabled using Trusted CA Certificate");
      }
    }
    else
    {
      request = new SOAPRequest(url, msg.getAction(), false);
    }
    
    if (headers != null)
    {
      for (var it = headers.keySet().iterator(); it.hasNext();)
      {
        var key = it.next();
        var value = headers.get(key);
        this.log.debug(key + "=" + value);
        request.getSoapHeaders().put(key, value);
      }
    }
    return new XML(request.post(msg.getSoapEnvelope(msg.getRequest()))).soap::Body;
  },
  
  /**
   * The default implementation of SSL between the IA and REST/SOAP endpoints
   * will allow for self-signed certificates. To implement more secure SSL
   * communication, an instance of an X509TrustManager should be created that
   * points to the keystore containing a signed certificate with the appropriate
   * credentials.
   *
   * The trusted certificate MUST be imported into the Java Keystore used by
   * the Integration Agent (conf/.keystore) using the credentials defined above.
   * In order for the trust chain to be enabled, the ACCEPT_ANY_CERTIFICATE
   * property above MUST be set to false                                         
   */
  getSSLSocketFactory: function()
  {
    if (this.SSL_SOCKET_FACTORY == null)
    {
      var sslContext = SSLContext.getInstance("TLS");
      var trustManagers = java.lang.reflect.Array.newInstance(TrustManager, 1);
      trustManagers[0] = new DefaultX509TrustManager(this.KEY_STORE, this.KEY_STORE_PASSWORD); 
      sslContext.init(null, trustManagers , null);
      this.SSL_SOCKET_FACTORY = new SSLSocketFactory(sslContext, SSLSocketFactory.STRICT_HOSTNAME_VERIFIER);
    }
    return this.SSL_SOCKET_FACTORY;
  }
});