importPackage(java.lang);
importPackage(java.util);
importPackage(java.io);
importClass(Packages.com.alarmpoint.integrationagent.config.IAConfigImpl);
importClass(Packages.com.alarmpoint.integrationagent.config.xml.IAConfigFileImpl);

/******************************************************************************/
/*        Start of xMatters Webservices Base Class Object Definition          */
/******************************************************************************/
var failedList = new ArrayList();
var successList = new ArrayList();
var warningList = new ArrayList();
var reprocessedList = new ArrayList();

var xmns = new Namespace("http://www.alarmpoint.com/webservices/schema");

var iaConfig = null;
var configFile = null;
var iaConfigFile = null;

var xMattersWS = WSUtil.extend({

  LOG_SOURCE: "xmattersws.js: ",

  XMATTERS_WSDL_DEFN: "integrationservices/caservicedesk/caservicedesk/soap/AlarmPoint-soapui-project.xml",
  XM_SOAP_VERSION: "1_1",
  SERVICE_DENIED: "SERVICE_DENIED",
  SUPERVISOR_INVALID: "SUPERVISOR_INVALID",
  UNKNOWN_SUPERVISOR: "UNKNOWN_SUPERVISOR",

  IA_CONFIG_FILE_PATH: "conf/IAConfig.xml",

  /**
   * Constructor
   */
  init: function()
  {
    this._super(); 

    if( iaConfig == null )
    {
      configFile = new File(this.IA_CONFIG_FILE_PATH);
      iaConfigFile = new IAConfigFileImpl(new FileReader(configFile));
      iaConfig = new IAConfigImpl(iaConfigFile, configFile.getParentFile(), configFile.toURI());
    }
    
    this.wsUser = iaConfigFile.getWSUser();
    this.wsPassword = iaConfig.getWsAuth().getPassword();
    this.wsCompany = iaConfigFile.getWSCompany();
    this.endPoint = iaConfigFile.getPrimaryServerUrls().next();
    
  },
  
  /**
   * Sends the request to the SOAP endpoint and returns the response
   * @param url EndPoint for the webservice request
   * @param msg SOAPMessage instance containing the SOAPEnvelope to post
   * @param objectType the type of object being synchronized (User, Coverage, Device, Group, Team)
   * @param objectName the name of the object being synchronized
   * @return SOAP Response Body
   */
  syncSendReceive: function(url, msg, objectType, objectName)
  {
    var response = this.sendReceive(url, msg);

    default xml namespace = xmns;
    var status = response..status;
    
    if (!"OK".equalsIgnoreCase(status))
    {
      var object = "user".equalsIgnoreCase(objectType) ? new User() : new Group();
      if (this.SUPERVISOR_INVALID.equalsIgnoreCase(status) || this.UNKNOWN_SUPERVISOR.equalsIgnoreCase(status))
      {
        var msg = objectType + ": " + objectName + " " + this.SUPERVISOR_INVALID + ". Setting supervisors to: " + object.supervisors;
        if (warningList.contains(msg))
        {
          warningList.remove(msg);
          failedList.add(objectType + ": " + objectName + " failed to synchronize: " + status);
        }
        else
        {
          reprocessedList.add(objectType + ": " + objectName);
          warningList.add(msg);
        }
      }
      else
      {
        failedList.add(objectType + ": " + objectName + " failed to synchronize: " + status);
      }
    }
    else
    {
      if (!reprocessedList.contains(objectType + ": " + objectName))
      {
        successList.add(objectType + ": " + objectName + " synchronized: " + status);
      }
    }
    return response;
  }
  
});
